# Hugo Weston 2022 hw876@bath.ac.uk

# Import other parts of this library:
import fyp_package.map
import fyp_package.sim

import numpy as np
import math
import acconeer.exptool as et

import scipy.io as scio # used to load matlab files

import serial
import time
import os
import pickle

import matplotlib.pyplot as plt


from scipy import signal

import copy


def reload_modules():
    import imp
    imp.reload(fyp_package.sim)
    imp.reload(fyp_package.map)
    imp.reload(fyp_package)


# Classes --------------------------------------------------------------------------------------
class linearMetadata:
    # This class contains all data for doing a linear run
    def __init__(self,directory,lens="unspecified",rangeInterval=[0.06,7.0],steps=range(0,100,10),serialPort="COM8",IP="192.168.1.75",sensorType="default"):
        # Set up class with lots of default parameters
        self.sensorType = sensorType
        self.A121_start_point = 0
        self.A121_num_points = 0
        self.lens = lens # Exists in name only, doesn't change anything in the code
        self.directory = directory
        self.IP =  IP
        self.update_rate = 30
        self.range = rangeInterval
        self.steps = steps
        self.HWAAS = 30 # Hardware Accelerated Average Samples
        self.stepmode = "stepmode_1"
        self.gain = 0.5
        self.serialPort = serialPort
        self.stepFactor = 0.20
        self.lowPassFilter = 0.5 # 0.5 = Turn filter off by default
        self.samples = 20 # Number of samples to take per data point
        self.stepPause = 0.5 # Time to pause inbetween steps
        self.datapointPause = 0.5 # time to wait after getting a set of data. Does not affect getting multiple samples, only affects calls to getnewdata()
        self.sensors = [1]
        self._saveMode = "samples"
        self.profile = ""

        if self.sensorType == "A121":
            self.create_depths()

    def create_depths(self):
        # Range for A121 is determined by a number of points, no acutal distance
        # Find number of points assocaited with chosen distance
        spacing = 2.5e-3

        self.A121_start_point = int(self.range[0]/spacing)

        current_position = self.A121_start_point * spacing
        num_points = 1
        while True:
            current_position += spacing
            if current_position > self.range[1]:
                break
            num_points += 1
        
        self.A121_num_points = num_points

    def numberOfSteps(self):
        # return the number of steps in the run 
        return len(self.steps)

    def stepsFromDistance(self, startPos, stopPos, samples):
        # Generate a list of positions (in steps) from a start and end position and a number of samples
        # Because of the conversion from distance to steps, stoppos and number of samples output may not exactly match input
        startFullSteps = startPos / self.stepFactor
        stopFullSteps = stopPos / self.stepFactor

        startSteps = startFullSteps * self.stepModeFactor()
        stopSteps = stopFullSteps * self.stepModeFactor()

        startSteps = int(startSteps)
        stopSteps = int(stopSteps)

        stepsBetweenSamples = int((stopSteps - startSteps) / samples)

        self.steps =  [*range(startSteps,stopSteps,stepsBetweenSamples)]

    def setStepMode(self, inputStepMode):
        if inputStepMode == 1:
            self.stepmode = "stepmode_1"
        elif inputStepMode == 2:
            self.stepmode = "stepmode_2"
        elif inputStepMode == 4:
            self.stepmode = "stepmode_4"
        elif inputStepMode == 8:
            self.stepmode = "stepmode_8"
        elif inputStepMode == 16:
            self.stepmode = "stepmode_16"
        elif inputStepMode == 32:
            self.stepmode = "stepmode_32"
        else:
            raise Exception("Incorrect step mode selection: '" + str(inputStepMode) + "'")
    
    def stepModeFactor(self):
        # The factor by which the chosen stepmode reduces the step
        return int(self.stepmode.replace("stepmode_",""))

    def alongTrackDirection(self):
        # Return a list of distances to take data at (mm)
        return np.array(self.steps) * self.stepFactor * (1/self.stepModeFactor())

    def setProfile(self,inputProfile):
        if self.sensorType == "A121":
            # No such thing as profile 1 or 5 for new sensors
            if inputProfile == 2:
                self._profile = et.configs.A121SparseIQConfig.Profile.PROFILE_2
            elif inputProfile == 3:
                self._profile = et.configs.A121SparseIQConfig.Profile.PROFILE_3
            elif inputProfile == 4:
                self._profile = et.configs.A121SparseIQConfig.Profile.PROFILE_4
            else:
                raise Exception("Incorrect profile selection: '" + str(inputProfile) + "' \n" + "Remember that there is no Profile 1 or 5 for the A121 Sensors" )
        else:
            if inputProfile == 1:
                self._profile = et.configs.IQServiceConfig.Profile.PROFILE_1
            elif inputProfile == 2:
                self._profile = et.configs.IQServiceConfig.Profile.PROFILE_2
            elif inputProfile == 3:
                self._profile = et.configs.IQServiceConfig.Profile.PROFILE_3
            elif inputProfile == 4:
                self._profile = et.configs.IQServiceConfig.Profile.PROFILE_4
            elif inputProfile == 5:
                self._profile = et.configs.IQServiceConfig.Profile.PROFILE_5
            else:
                raise Exception("Incorrect profile selection: '" + str(inputProfile) + "'")
        self.profile = "Profile " + str(inputProfile)

    def savetxt(self):
        # Save the run info to a file in a human-readable manner
        touchDir(self.directory) # create directory if it doesn't already exist
        selfDict = vars(self)

        f = open("./" + self.directory + "/run_info_readable.txt","w")
        f.write("Data saved for reading only.\n")
        f.write("TIMESTAMP of file creation (not time of run):\n")
        f.write(str(time.ctime(time.time())))
        f.write("\n")

        f.write("Calculated Information:")
        f.write("\n")
        f.write("Number of datapoints: " + str(len(self.steps)))
        f.write("\n")
        f.write("Min Distance: " + str(self.steps[0] * self.stepFactor / self.stepModeFactor()) + " mm")
        f.write("\n")
        f.write("Max Distance: " + str(self.steps[-1] * self.stepFactor / self.stepModeFactor()) + " mm")
        f.write("\n")

        f.write("Variable Information: \n")
        for key in selfDict:
            f.write(str(key) + ": " +  str(selfDict[key]))
            f.write("\n")
        f.close()
    
    

    def saveInfo(self):
        touchDir(self.directory) # create directory if it doesn't already exist
        #Replacement for the old saveInfo. Does not use pickle and therefore does not require consistent library/module/class names
        if os.path.exists(self.directory + "/run_info.hugo"):
            print("Warning. File '" + str(self.directory + "/run_info.p") + "' already exists")
            userCheck = input("overwrite? (y/n) ")
            if userCheck.upper() != "Y":
                raise Exception("User chose not to overwrite Data")

        
        # convert to dictionary
        infoToSave = vars(self)
        f = open("./" + self.directory + "/run_info.hugo","w")
        for key in infoToSave.keys():
            f.write(key)
            f.write(";")
            f.write(str(infoToSave[key]))
            f.write("\n")
        f.close()

    def saveInfoOld(self):
        # Save the run info to a file so it can be loaded by a computer at a later date
        if os.path.exists(self.directory + "/run_info.p"):
            print("Warning. File '" + str(self.directory + "/run_info.p") + "' already exists")
            userCheck = input("overwrite? (y/n) ")
            if userCheck.upper() != "Y":
                raise Exception("User chose not to overwrite Data")


        touchDir(self.directory) # create directory if it doesn't already exist
        f = open("./" + self.directory + "/run_info.p","wb") #.p = pickle file
        pickle.dump(self, f)
        f.close()

# Signal & Data Processing ----------------------------------------------------------------------------

def average(inputVector): # averages a series of vectors
    # Designed for input vector to be a 1D array of 1D vectors
    # Will average for each item in the list 

    # Used to average multiple samples of data
    avgVector = inputVector[0]
    for i in range(1,len(inputVector)):
        avgVector = avgVector + inputVector[i]
    avgVector = avgVector / len(inputVector)
    return avgVector

# Filtering functions are commonly used, but were taken from following sources:
# https://medium.com/analytics-vidhya/how-to-filter-noise-with-a-low-pass-filter-python-885223e5e9b7
# https://www.delftstack.com/howto/python/low-pass-filter-python/
def butter_lowpass(cutoff, fs, order=5):
    nyq = 0.5 * fs
    normal_cutoff = cutoff / nyq
    b, a = signal.butter(order, normal_cutoff, btype='low', analog=False)
    return b, a

def butter_lowpass_filter(data, cutoff, fs, order=5):
    # data = vector of data
    # cuttoff = cuttoff frequency to filter out
    # fs = sampling frequency
    b, a = butter_lowpass(cutoff, fs, order=order)
    y = signal.lfilter(b, a, data)
    return y


def ncrossCorr(dataset1, dataset2,mode="normal"):
    # Return the normalised cross correlation between dataset1 and dataset2
    
    if mode == "phase":
        # Phase only cross correlation
        for i in range(len(dataset1)):
            # Normalise the magnitude of the vectors without modifying the angle
            # This means the only differences between datapoints is the phase
            if np.abs(dataset1[i]) != 0:
                dataset1[i] = dataset1[i] / np.abs(dataset1[i])
            if np.abs(dataset2[i]) != 0:
                dataset2[i] = dataset2[i] / np.abs(dataset2[i])

    acorr1 = signal.correlate(dataset1,dataset1,mode="full") # auto correlation of dataset 1
    acorr2 = signal.correlate(dataset2,dataset2,mode="full") # auto correlation of dataset 2

    # Cross correlation:
    ccorr = signal.correlate(dataset1,dataset2,mode="full")

    # Normalise the Cross correlation
    ccorr = ccorr / np.sqrt(np.abs(max(np.abs(acorr1))) * np.abs(max(np.abs(acorr2))))
    
    return ccorr

def peak_value(dataset1,dataset2,mode="both"):
    cross_correlation = ncrossCorr(dataset1,dataset2,mode)
    #middleindex = int((len(cross_correlation) - 1)/2)
    #print(middleindex)
    #return np.abs(cross_correlation[middleindex])
    return max(np.abs(cross_correlation))


# Gathering Data ----------------------------------------------------------------------------------

def getDepths(runInfo):
    # Return the depths bins for a input sensor setup
    [client,config] = setupClient(runInfo)
    session_info = client.setup_session(config)

    if runInfo.sensorType == "A121":
        depths = et.utils.approx_r3_range_m(config)[0]
    else:
        depths = et.utils.get_range_depths(config, session_info)
    return depths

def getSamplingFreq(depths):
    # Calculate sample frequency:
    c = 299792458 # speed of light to a unecessary level of accuracy
    rangeBin = (depths[-1] - depths[0]) / len(depths)
    ts = (2*rangeBin)/c # time = distance/speed
    sr = 1/ts # sampling rate
    return sr    

def getNewData(runInfo):

    [client,config] = setupClient(runInfo)

    client.connect()

    session_info = client.setup_session(config)
    #depths = et.utils.get_range_depths(config, session_info)
    #print("Session info:\n", session_info, "\n")
    client.start_session()
    dataset = [] # data to return
    for i in range(runInfo.samples):
        dataset.append(client.get_next()[1])
    
    client.stop_session()
    time.sleep(0.1)
    client.disconnect()
    
    # pause after getting data
    # Sometimes the pi is still connected afterwards. Pause to give it time to disconnect before reconnecting 
    time.sleep(runInfo.datapointPause)
    return dataset

def setupClient(runInfo):

    client = et.SocketClient(runInfo.IP)

    if runInfo.sensorType == "A121":
        config = et.configs.A121SparseIQConfig()
        config.sweeps_per_frame = 1
        config.start_point = runInfo.A121_start_point
        config.step_length = 1
        config.num_points = runInfo.A121_num_points

        config.frame_rate = runInfo.update_rate
        config.receiver_gain  = runInfo.gain
        config.hwaas = runInfo.HWAAS
    else:
        config = et.IQServiceConfig()
        config.range_interval = runInfo.range
        config.asynchronous_measurement = False # slower, but prevents interference between sensors
        config.depth_lowpass_cutoff_ratio = runInfo.lowPassFilter
        config.noise_level_normalization = False # added 27/02/2022
        config.update_rate = runInfo.update_rate
        config.gain = runInfo.gain
        config.hw_accelerated_average_samples = runInfo.HWAAS

    config.sensor = runInfo.sensors
    
    config.profile = runInfo._profile
    
    

    return [client,config]

def LinearRun(listOfRuns):
    # Do a run and get data at points along the run

    # Metadata about the run is saved when this function is called - it is not requried to save this seperately

    # Will do separate runs in parallel. Therefore all runs must have the same:
    # - Lens
    # - Steps/Number of dataPoints/Start and end position
    # - Scene

    # Serial connection is one-way, so need to wait between sending signals
    arduino=serial.Serial(listOfRuns[0].serialPort, 9600)
    time.sleep(0.5)
    print("Arudino Connection Established")
    arduino.write("0".encode()) # write it to position 0 (should be current position)
    time.sleep(1)
    arduino.write(listOfRuns[0].stepmode.encode()) # Transmit desired microstep mode
    time.sleep(1)
    dataset = []

    # Save metadata
    for i in range(len(listOfRuns)): # loop through differnt runs
        # Create directory if it doesn't already exist
        touchDir(listOfRuns[i].directory)
        # Save setup
        #listOfRuns[i].saveInfo()
        #listOfRuns[i].savetxt()
        
        # Save depths
        depths = getDepths(listOfRuns[i])
        np.save("./" + listOfRuns[i].directory + "/depths",depths)

    for i in range(len(listOfRuns)):
        dataset.append([]) # add new empty list for data for each run
    
    for i in range(len(listOfRuns[0].steps)):
        # Loop through step positions
        arduino.write(str(listOfRuns[0].steps[i]).encode())
        time.sleep(listOfRuns[0].stepPause)
        print(str(i+1) + "/" + str(len(listOfRuns[0].steps)))

        for j in range(len(listOfRuns)):
            #remove later:
            print("Profile: " + str(listOfRuns[j]._profile) + " HWAAS: " + str(listOfRuns[j].HWAAS))
            # Runs loop
            data = getNewData(listOfRuns[j])
            """
            if len(listOfRuns[j]) == 1:
                data = [data]
            """
            dataset[j].append(data)

            runInfo = listOfRuns[j]

            # save data
            if runInfo._saveMode == "samples" or runInfo._saveMode == "both":
                touchDir("./" + runInfo.directory + "/data/samples/")

                # Save the full sampled data
                # modified DATASET is the list of samples (not HWAAS)
                if len(runInfo.sensors) == 1: # Store the data in a 1-item list if only using 1 sensor
                    if runInfo.sensorType == "A121":
                        modifiedAverage = average(dataset[j][i])
                        modifiedDataset = dataset[j][i]
                    else:
                        modifiedAverage = [average(dataset[j][i])]
                        modifiedDataset = [dataset[j][i]]
                else: # Data for multiple sensors is stored in a list by default
                    modifiedDataset = dataset[j][i]
                    modifiedAverage = average(dataset[j][i])

                np.save("./" + runInfo.directory + "/data/samples/samples datapoint " + str(i),
                modifiedDataset) #dtype=object tells numpy that this is a complex data structure

            if runInfo._saveMode == "average" or runInfo._saveMode == "both":
                touchDir("./" + runInfo.directory + "/data/average/")
                np.save("./" + runInfo.directory + "/data/average/average datapoint " + str(i),
                modifiedAverage) #dtype=object tells numpy that this is a complex data structure
    
    # Reset Position back to 0
    arduino.write(str(listOfRuns[0].steps[0]).encode())


# Loading and Saving Data ---------------------------------------------------------------------

def loadRunInfoOld(directory):
        # Load a previously saved run setup

        f = open(directory + "/run_info.p","rb")
        data = pickle.load(f)
        f.close()
        return data

def loadRunInfo(directory):
    s = linearMetadata(directory) # will initially set the wrong directory, this will be overwritten
    f = open(directory + "/run_info.hugo","r")
    for line in f.readlines():
        [key,value] = line.split(";")
        value = value[:-1] # remove newline char
        #print(str(key) + ", " + str(value))

        # Why doesn't python have a switch?!?
        if key == "lens":
            s.lens = value
        elif key == "directory":
            s.directory = value
        elif key == "IP":
            s.IP = value
        elif key == "update_rate":
            s.update_rate = int(value)
        elif key == "range":
            listValues = value[1:-1].split(",")
            if len(listValues) > 2:
                raise Exception("Unexpected error. Range list longer than 2")
            s.range = [float(listValues[0]),float(listValues[1])]
        elif key == "steps":
            listValues = value[1:-1].split(",")
            s.steps = []
            for listValue in listValues:
                s.steps.append(int(listValue))
        elif key == "HWAAS":
            s.HWAAS = int(value)
        elif key == "stepmode":
            s.stepmode = value
        elif key == "gain":
            s.gain = float(value)
        elif key == "serialPort":
            s.serialPort = value
        elif key == "stepFactor":
            s.stepFactor = float(value)
        elif key == "lowPassFilter":
            s.lowPassFilter = float(value)
        elif key == "samples":
            s.samples = int(value)
        elif key == "stepPause":
            s.stepPause = float(value)
        elif key == "sensors":
            s.sensors = []
            listValues = value[1:-1].split(",")
            for listValue in listValues:
                s.sensors.append(int(listValue))
        elif key == "_saveMode":
            s._saveMode = value
        elif key == "profile":
            profileValue = value.replace("Profile ","")
            #s.setProfile(int(profileValue))
            s.profile = str(int(profileValue))
        elif key == "_profile":
            # do nothing
            pass
        elif key == "sensorType":
            s.sensorType = value
        elif key == "A121_start_point":
            s.A121_start_point = value
        elif key == "A121_num_points":
            s.A121_num_points = value
        elif key == "datapointPause":
            s.datapointPause = value
        else:
            raise Exception("Unexpected metadata: " + str(key))
    f.close()
    return s

def loadRunsInfo(fileListDirectory,directoryPrefix="."):
    f = open(directoryPrefix + "/" + fileListDirectory + "/list_of_runs.txt","r")
    runsInfo = []

    for line in f.readlines():
        if directoryPrefix == ".":
            run = loadRunInfo("./" + line[:-1]) # remove the \n char 
        else:
            run = loadRunInfo(directoryPrefix + "/" + line[:-1]) # remove the \n char 
            run.directory = directoryPrefix + "/" + line[:-1] # set to new directory
        runsInfo.append(run)
    f.close

    return runsInfo
    
def saveListOfRuns(listOfRuns,directory):
    touchDir(directory)
    f = open("./" + directory + "/list_of_runs.txt","w")
    for run in listOfRuns:
        run.saveInfo()
        run.savetxt()
        f.write(run.directory)
        f.write("\n")
    f.close()


def loadDataFromListOfRuns(listOfRuns,mode="average"):
    data = []
    for run in listOfRuns:
        data.append(loadDataFromRun(run,mode))
    return data

def loadDataFromRun(runInfo,mode="average"):
    # Load data saved in the directory in runInfo
    # mode determines what data to load: samples, averages, or both
    # loading only averages is much faster, so that is recommended unless you really need samples

    class dataSet:
        def __init__(self):
            self.samples = []
            self.average = []
    
    data = [] # variable to return data in
    for i in range(len(runInfo.sensors)): # append item for each sensor data to load
        data.append(dataSet())

    if mode == "samples" or mode == "both":
        for i in range(len(runInfo.steps)):
            fileData = np.load("./" + runInfo.directory + "/data/samples/" + "samples datapoint " + str(i) + ".npy",allow_pickle=True)

            # need to unpack values if only using 1 sensor
            if len(runInfo.sensors) == 1 and runInfo.sensorType != "A121":
                fileData = fileData[0]
                samples = []

                #if runInfo.sensorType == "A121" and len(fileData) == 1: # new sensors store data differently. need to unpack data
                #    fileData = fileData[0]

                for k in range(runInfo.samples):
                    samples.append(fileData[k])
                data[0].samples.append(samples)
            else:
                for j in range(len(runInfo.sensors)):
                    samples = []
                    for k in range(runInfo.samples):
                        samples.append(fileData[k][j])
                    data[j].samples.append(samples) #data[j].samples.append(fileData[j])
    
    if mode == "average" or mode == "both":
        for i in range(len(runInfo.steps)):
            fileData = np.load("./" + runInfo.directory + "/data/average/" + "average datapoint " + str(i) + ".npy",allow_pickle=True)
            for j in range(len(runInfo.sensors)):
                data[j].average.append(fileData[j]) #data[j].samples.append(fileData[j])

    if runInfo._saveMode != mode:
        print("FYI: Data loading mode '" + str(mode) + "' does not match data collection mode '" + str(runInfo._saveMode) + "'")

    for i in range(len(runInfo.sensors)):
        data[i].length = max(len(data[i].samples),len(data[i].average))

        data[i].depths = np.load("./" + runInfo.directory + "/depths.npy",allow_pickle=True)
    return data


def loadMatlabData(filepath):
    # Load esitmated correlation profile from a matlab file

    
    fileData = scio.loadmat(filepath)# import the expected profile variation
    tempDis = fileData["surgeVec"][0] # unpack data
    tempDis = list(tempDis) # unpack data

    # Each item in the array is stored in a 1-element array
    # Need to unpack these values
    expectedDis = []
    expectedRho = []
    for i in range(len(fileData["rho"])):
        if not(np.isnan(fileData["rho"][i][0])):
            expectedRho.append(fileData["rho"][i][0]) # 0th item in 1-element array
            expectedDis.append(tempDis[i])

    # The data from matlab is single-sided
    # Make it double sided
    doubleSideRho = []
    doubleSideDis = []

    for i in range(len(expectedRho) - 1):
        doubleSideRho.insert(0,expectedRho[i+1])
        doubleSideDis.insert(0,-expectedDis[i+1])

    expectedRho = np.array(doubleSideRho + expectedRho)
    expectedDis = np.array(doubleSideDis + expectedDis)

    return [expectedDis, expectedRho]

# Misc --------------------------------------------------------------------------------------

def touchDir(directory):
    # directory is a (relative) directory path using "/" as seperators
    # Will create a directory path if it doesn't already exist
    folderNames = directory.split("/") # get name of each folder

    # Remove first character if it references the current directory
    if folderNames[0] == ".":
        folderNames.pop(0)

    currentDirectory = "./"
    for folderName in folderNames:
        if not(os.path.isdir(currentDirectory + folderName + "/")):
            os.mkdir(currentDirectory + folderName + "/")
        currentDirectory = currentDirectory + folderName + "/"


# Data Analysis Functions  -----------------------------------------------------------------------------

def compareSignals(dataset1, dataset2, depths, saveDir = "",filename="",pltLabels = ["",""],title="",correlationMode="normal"):
    autoCorrelation1 = ncrossCorr(dataset1,dataset1,mode=correlationMode)
    autoCorrelation2 = ncrossCorr(dataset2,dataset2,mode=correlationMode)
    crossCorrelation = ncrossCorr(dataset1,dataset2,mode=correlationMode)

    fig, axs = plt.subplots(5,1,figsize=(20,25))
    fig.subplots_adjust(hspace=0.45) # give more hspace between plots to make axes labels visible


    fig.suptitle(title)
    axs[0].plot(depths,np.abs(dataset1), label=pltLabels[0])
    axs[0].plot(depths,np.abs(dataset2), label=pltLabels[1])
    axs[0].set_xlabel("Distance (m)")
    axs[0].set_ylabel("Amplitude")
    axs[0].set_title("Amplitude vs Distance")
    axs[0].grid(visible=True)
    axs[0].legend()

    axs[1].plot(depths,np.angle(dataset1), label=pltLabels[0])
    axs[1].plot(depths,np.angle(dataset2), label=pltLabels[1])
    axs[1].set_xlabel("Distance (m)")
    axs[1].set_ylabel("Phase")
    axs[1].set_title("Phase vs Distance")
    axs[1].grid(visible=True)
    axs[1].legend()

    distance = depths[-1] - depths[0]
    tmp1 = np.linspace(-distance,distance,len(autoCorrelation1)) # temporary x vector

    axs[2].plot(tmp1,np.abs(autoCorrelation1), label=pltLabels[0] + " AutoCorrelation")
    axs[2].set_xlabel("lag (m)")
    axs[2].set_ylabel("Correlation Magnitude")
    axs[2].set_title("Datum AutoCorrelation")
    axs[2].grid(visible=True)
    axs[2].legend()


    axs[3].plot(tmp1,np.abs(autoCorrelation2), label=pltLabels[1] + "AutoCorrelation")
    axs[3].set_xlabel("Distance (m)")
    axs[3].set_ylabel("Correlation Magnitude")
    axs[3].set_title("Sample AutoCorrelation")
    axs[3].grid(visible=True)

    axs[4].plot(tmp1,np.abs(crossCorrelation), label="Cross Correlation")
    axs[4].set_xlabel("Distance (m)")
    axs[4].set_ylabel("Correlation Magnitude")
    axs[4].set_title("Datum-Sample Cross Correlation. Max: " + str(max(np.abs(crossCorrelation))))
    axs[4].grid(visible=True)
    axs[4].set_ylim(0,1)

    if saveDir != "":
        touchDir(saveDir + "/saveDataPointFigures/")
        plt.savefig(saveDir + "/saveDataPointFigures/" + filename + ".png",dpi=300)
        plt.close("all")

    

def saveDataPointFigures(runInfo,data,depths,datum):
    # For a given dataset, this will save a figure for each averaged datapoint
    # data = list of AVERAGED datapoint values. Each datapoint should have a single vector in it
    # Datum = usually just the first item in the data eg: data[0], but could be anything (like the centre point)

    print("Saving figures for each datapoint, this may take a while...")
    

    # loop through data points
    for i in range(len(data)): # dataPOINT loop
        print("Datapoint " + str(i+1) + "/" + str(len(data)))

        autoCorrelation1 = ncrossCorr(datum,datum)
        autoCorrelation2 = ncrossCorr(data[i],data[i])
        #crossCorrelation = correlationVector
        crossCorrelation = ncrossCorr(datum,data[i])

        fig1, axs = plt.subplots(5,1,figsize=(20,25))
        fig1.subplots_adjust(hspace=0.45) # give more hspace between plots to make axes labels visible
        fig1.suptitle("Distance: " + str(runInfo.alongTrackDirection()[i] * runInfo.stepFactor * (1/runInfo.stepModeFactor())) \
                          + "mm")
        axs[0].plot(depths,np.abs(datum), label="Datum (average)")
        axs[0].plot(depths,np.abs(data[i]), label="Sample (average)")
        axs[0].set_xlabel("Distance (m)")
        axs[0].set_ylabel("Amplitude")
        axs[0].set_title("Amplitude vs Distance")
        axs[0].grid(visible=True)
        axs[0].legend()

        axs[1].plot(depths,np.angle(datum), label="Datum (average)")
        axs[1].plot(depths,np.angle(data[i]), label="Sample (average)")
        axs[1].set_xlabel("Distance (m)")
        axs[1].set_ylabel("Phase")
        axs[1].set_title("Phase vs Distance")
        axs[1].grid(visible=True)
        axs[1].legend()

        tmp1 = np.linspace(-1,1,len(autoCorrelation1)) # temporary x vector

        axs[2].plot(tmp1,np.abs(autoCorrelation1), label="Datum AutoCorrelation")
        axs[2].set_xlabel("Distance (m)")
        axs[2].set_ylabel("Correlation Magnitude")
        axs[2].set_title("Datum AutoCorrelation")
        axs[2].grid(visible=True)
        axs[2].legend()


        axs[3].plot(tmp1,np.abs(autoCorrelation2), label="Sample AutoCorrelation")
        axs[3].set_xlabel("Distance (m)")
        axs[3].set_ylabel("Correlation Magnitude")
        axs[3].set_title("Sample AutoCorrelation")
        axs[3].grid(visible=True)

        axs[4].plot(tmp1,np.abs(crossCorrelation), label="Datum-Sample Cross Correlation")
        axs[4].set_xlabel("Distance (m)")
        axs[4].set_ylabel("Correlation Magnitude")
        axs[4].set_title("Datum-Sample Cross Correlation. Max: " + str(max(np.abs(crossCorrelation))))
        axs[4].grid(visible=True)
        axs[4].set_ylim(0,1)

        touchDir(runInfo.directory + "/saveDataPointFigures/")
        plt.savefig(runInfo.directory + "/saveDataPointFigures/" + str(i) + ".png",dpi=300)
        plt.close("all")
               
    
def saveDataSampleFigures(runInfo,data,depths,datum):
    # HIGHLY UNRECOMMENDED

    # Generates a plot for each data sample
    # Will take ages to run
    # Will create gigabytes worth of figures

    touchDir(runInfo.directory + "/saveDataPointFigures/")

    for i in range(len(data)):
        touchDir(runInfo.directory + "/saveDataPointFigures/" + str(i) + "/")
        for j in range(len(data[i])):
            acorr1 = ncrossCorr(datum,datum)
            ccorr = ncrossCorr(datum,data[i][j])
            acorr2 = ncrossCorr(data[i][j],data[i][j])

            fig0, axs = plt.subplots(3,2,figsize=(20,14))
            fig0.suptitle("Distance: " + str(runInfo.alongTrackDirection()[i] * runInfo.stepFactor * (1/runInfo.stepModeFactor())) \
                          + "mm")

            tmp1 = np.linspace(-3.14,3.14,len(acorr1))
            axs[0,0].plot(depths,np.abs(datum),color= 'red')
            axs[0,0].set_title("Datum Signal")
            axs[0,0].set_xlabel("Distance")
            axs[0,0].set_ylabel("Amplitude",color= 'red')
            axs[0,0].tick_params(axis ='y', labelcolor = 'red') 

            tmpaxs = axs[0,0].twinx()
            tmpaxs.plot(depths,np.angle(datum),color= 'blue')
            tmpaxs.set_ylabel("Phase",color= 'blue')
            tmpaxs.tick_params(axis ='y', labelcolor = 'blue') 


            axs[0,1].plot(depths,np.abs(data[i][j]),color= 'red')
            axs[0,1].set_title("Current Signal")
            axs[0,1].set_xlabel("Distance")
            axs[0,1].set_ylabel("Amplitude",color= 'red')
            axs[0,1].tick_params(axis ='y', labelcolor = 'red') 

            tmpaxs = axs[0,1].twinx()
            tmpaxs.plot(depths,np.angle(data[i][j]),color= 'blue')
            tmpaxs.set_ylabel("Phase",color= 'blue')
            tmpaxs.tick_params(axis ='y', labelcolor = 'blue') 

            axs[1,0].plot(tmp1,np.abs(acorr1))
            axs[1,0].set_title("Autocorrelation of Datum (Amplitude)")

            axs[1,1].plot(tmp1,np.abs(acorr2))
            axs[1,1].set_title("Autocorrelation of Current (Amplitude)")
                    
            axs[2,1].plot(tmp1,np.abs(ccorr))
            axs[2,1].set_title("Cross Correlation (Amplitude): Peak Value: " + str(np.abs(max(np.abs(ccorr)))))
            axs[2,1].set_ylim(0,1)
            
            plt.savefig(runInfo.directory + "/saveDataPointFigures/" + str(i) + "/" + str(j) + ".png")
            plt.close("all")
