# Mapping code for my fyp
# Hugo Weston 2022 hw876@bath.ac.uk

import numpy as np
import copy

def normaliseArray(arr):
    #normalise 2D matrix
    #Also convert it to only amplitude
    
    #get max value:
    currentMax = np.abs(arr[0][0])
    for x in range(len(arr)):
        for y in range(len(arr[0])):
            if np.abs(arr[x][y]) > currentMax:
                currentMax = np.abs(arr[x][y])

    if currentMax == 0:
        raise Exception("Found that the maximum value in the array is 0. Cannot normalise array")
    
    for x in range(len(arr)):
        for y in range(len(arr[0])):
            arr[x][y] = np.abs(arr[x][y]) / currentMax
    return arr
        
def averageSignalArray(signalArray,countArray):
    for y in range(len(signalArray)):
        for x in range(len(signalArray[0])):
            #print("x = " + str(x) + " y = " + str(y))
            if countArray[y][x] > 0: # Don't mess with anything if no signal is present in this pixel
             signalArray[y][x] = signalArray[y][x] / countArray[y][x]
    return signalArray


def create_canvas(sizex,sizey,canvasType="signal"):
    # canvasType = signal returns a 2D array of zeros
    # canvasType = image returns a 2D array of [0.0, 0.0, 0.0] lists (rgb)

    canvas = []
    for i in range(sizey):
        canvas.append([])
        for j in range(sizex):
            if canvasType == "signal":
                canvas[i].append(0.0)
            elif canvasType == "image":
                canvas[i].append([0.0,0.0,0.0]) # must be float
            else:
                raise Exception("Incorrect canvas type selection: '" + str(canvasType) + "'")
    if canvasType == "signal":
        return np.array(canvas,dtype="cfloat")
    elif canvasType == "image":
        return canvas
    else:
        raise Exception("Incorrect canvas type selection: '" + str(canvasType) + "'")   


def indexMapSwath(signalLength,halfSwathWidthDeg):
    # This function maps the datapoints from the 1-D signal array the 2-D image array

    # Half Swath Width is half the angle of the projected signal
    # Signal length is the length of the signal vector
    halfSwathWidthRad = halfSwathWidthDeg * np.pi/180
    
    arrayHeight = signalLength
    arrayWidth = int(np.ceil(2 * signalLength * np.sin(halfSwathWidthRad)))

    # need the width to the odd
    if arrayWidth % 2 == 0:
        arrayWidth += 1
    
    
    imageArray = create_canvas(arrayWidth,arrayHeight,canvasType="signal")

    sensorXPos = int(np.ceil(arrayWidth/2))

    class signalMap():
        def __init__(self,arrayWidth,arrayHeight):
            self.targetX = []
            self.targetY = []
            self.signal = []
            self.arrayWidth = arrayWidth
            self.arrayHeight = arrayHeight
            self.sensorXPos = sensorXPos
            self.weightings_basic = []

    output_signalMap = signalMap(arrayWidth,arrayHeight)
    
    # iterate through array
    for x in range(imageArray.shape[1]):
        for y in range(imageArray.shape[0]):
            xDist = sensorXPos - x
            yDist = y


            verticalAngle = np.arctan2(xDist,yDist) # angle between vector and vertical
            
            # Check the angle is within requested
            if np.abs(verticalAngle) <= halfSwathWidthRad:
                r = (xDist**2 + yDist**2)**0.5 # pythagorean theorem
                # r is the distance in samples between [x,y] and the array position


                # Check that you can actually populate the data
                if int(r) < signalLength:
                    output_signalMap.targetX.append(x)
                    output_signalMap.targetY.append(y)
                    output_signalMap.signal.append(int(r)) # round to the nearest whole number
                    output_signalMap.weightings_basic.append(np.exp(-(5*verticalAngle)**2))
                
    output_signalMap.targetX = np.array(output_signalMap.targetX)
    output_signalMap.targetY = np.array(output_signalMap.targetY)
    return output_signalMap

def processSARImage(run,data,userResolution,halfSwathWidth,filterVal=0):
    # Calculate Resolutions:
    alongResolution = (run.alongTrackDirection()[-1] - run.alongTrackDirection()[0]) / len(run.alongTrackDirection()) # mm per sample taken

    #acrossResolution = 1000 * (run.range[1] - run.range[0]) / len(data.depths) # mm per sample (across track)

    signalLength_mm = data.depths[-1] * 1000 # max depths (mm)
    signalLength_sa = int(signalLength_mm /userResolution) # max depths (sampes)

    extraWidth = np.sin(halfSwathWidth * np.pi/180) * signalLength_sa # this is the width caused by the swath expanding outwards

    # + 10 as a tolerance as the indexing may go to above the width
    imageWidth = int(len(data.average) * alongResolution * (1/userResolution)  + 2*extraWidth + 10)   # convert to have the same sample resolution as the range 


    signalArray = create_canvas(imageWidth,signalLength_sa,canvasType="signal")
    countArray = create_canvas(imageWidth, signalLength_sa,canvasType="signal")
    # Low Pass Filter data
    c = 299792458
    D = (data.depths[-1] - data.depths[0])/len(data.depths)
    d = 2*D/c

    print("Creating Image Index Map")
    indexMap = indexMapSwath(signalLength_sa,halfSwathWidth) # map signal to 2D array

    print("Creating Signal Index Map")
    pointFrom = []
    pointTo = []
    for i in range(signalLength_sa):
        #print(str(i))
        pos = i * userResolution / 1000
        for j in range(len(data.depths) - 1):
            if data.depths[j] < pos and data.depths[j+1] >= pos:
                pointFrom.append(j)
                pointTo.append(i)
    
    for i in range(len(data.average)):
        print(str(i+1) + "/" + str(len(data.average))) # Progress Bar

        signal = data.average[i]
        
        # Change comment to add/remove filtering
        #signal = fyp.butter_lowpass_filter(imageData.average[i],10e9,1/d)
        # remove close signal data
        #signal[0:250] = 0
        
        modifiedSignal = np.zeros(signalLength_sa,dtype="cfloat")
        modifiedSignal[pointTo] = signal[pointFrom] # insert signal

        #modifiedSignal = np.abs(modifiedSignal) # Amplitude only map (not SAR)
        
        xOffset = int(run.alongTrackDirection()[i]/userResolution)
        signalArray[indexMap.targetY,indexMap.targetX + xOffset] += modifiedSignal[indexMap.signal] * indexMap.weightings_basic
        countArray[indexMap.targetY,indexMap.targetX + xOffset] += 1

    croppedArray = copy.deepcopy(signalArray[:,int(extraWidth):imageWidth-int(extraWidth)])
    return croppedArray
