# Simulation code for my fyp
# Hugo Weston 2022 hw876@bath.ac.uk

import scipy.io as scio # used to load matlab files

import numpy as np

import os

# class to store a simulated sensor array
class simSensorArray ():
    def __init__(self,sensor_separations):
        self.sensors = []
        for sensor_separation in sensor_separations:
            self.sensors.append(simSensor(sensor_separation))
        #self.estimate_mode = ""
        #self.correlation_mode = ""



    def inter_correlation(self,distance_vector, base_sensor, target_sensor, target_sensor_position, mode="ideal"):
        # position is how far the target sensor has moved!
        if mode == "ideal":
            int_correlation = self.sensors[base_sensor].ideal_correlation(distance_vector, target_sensor_position + self.sensors[target_sensor].separation)
        elif mode == "real":
            int_correlation = self.sensors[base_sensor].real_correlation(distance_vector, target_sensor_position + self.sensors[target_sensor].separation)
        elif mode == "estimated":
            int_correlation = self.sensors[base_sensor].estimated_correlation(distance_vector, target_sensor_position + self.sensors[target_sensor].separation)
        else:
            raise Exception("Incorrect mode: " + str(mode))
        return int_correlation
    
    def inter_displacement_estimate(self,distance_vector,time_vector,base_sensor, target_sensor, target_sensor_position, cor_mode="ideal",est_mode="default"):
        
        int_correlation = self.inter_correlation(distance_vector, base_sensor, target_sensor, target_sensor_position,mode=cor_mode)
        [estTime,estDist] = self.sensors[base_sensor].estimate_displacement(int_correlation,time_vector,vals=est_mode) #gauss_parameters_mitigated


        distance_between_sensors = self.sensors[target_sensor].separation - self.sensors[base_sensor].separation
        estDist = estDist + distance_between_sensors + target_sensor_position
        return [estTime,estDist]

    def inter_distance_estimate(self,distance_vector,time_vector,base_sensor, target_sensor, target_sensor_position):
        [estTime,estDist] = self.inter_displacement_estimate(distance_vector,time_vector,base_sensor, target_sensor, target_sensor_position)
        estDist = estDist + self.sensors[base_sensor].separation
        return [estTime,estDist]
        

                

    
    

# class to store a simulated sensor
class simSensor():
    def __init__(self,sensor_separation):
        self.separation = sensor_separation
        self.estimatedDistance = []
        self.estimatedTime = []
        
    # These are all ways of getting a correlation profile
    def ideal_correlation(self,distanceVector,datum): # datum is a distance
        idealCorrelaion = gauss(distanceVector - datum + self.separation)
        #unused (for now)
        return idealCorrelaion
    
    def real_correlation(self,distanceVector,datum):
        idealCorrelaion = self.ideal_correlation(distanceVector,datum)
        #noise = 0.05 * np.random.normal(size=(len(distanceVector)))
        noise = 0.01 * np.random.normal(size=(len(distanceVector)))
        realCorrelation = 0.909090909 * 0.75 * 0.9 * idealCorrelaion + noise + 0.1
        return realCorrelation
    
    def custom_correlation():
        return

    def estimated_correlation(self,distanceVector,datum,vals="gauss_parameters_mitigated"):
        return gauss(distanceVector - datum + self.separation,vals=vals)
    

    # Estimating displacement/distance DOES NOT TAKE INTO ACOUNT THE DATUM POSITION
    # In your code you MUST REMEMBER TO ADD THE DATUM BACK ON
    def estimate_displacement(self,correlation,timeVector,vals="default",minCorrelation=0.5):
        # Alternatively use "gauss_parameters_mitigated"
        
        
        estimated_distance_1 = np.array([]) # store low estimate values
        estimated_distance_2 = np.array([]) # store high estimate values
        estimatedTime = np.array([])
        #estimated_time = np.zeros(len(correlation)) # time vector corresponding to estimated distances
        
        for i in range(len(correlation)):
            if correlation[i] > minCorrelation:
                estimated_distance_1 = np.append(estimated_distance_1, invGauss(correlation[i],vals=vals)[0])
                estimated_distance_2 = np.append(estimated_distance_2, invGauss(correlation[i],vals=vals)[1])
                estimatedTime = np.append(estimatedTime, timeVector[i])
        return [np.concatenate((estimatedTime, estimatedTime)), np.concatenate((estimated_distance_1, estimated_distance_2))]
    
    def estimate_distance(self, correlation, timeVector, vals="default"):
        [estTime, estDist] = self.estimate_displacement(correlation, timeVector,vals=vals)
        estDist = estDist + self.separation
        return [estTime, estDist]
    
        

def getRsquared(data, fittedCurve):
    # Find r squared (coefficient of determination) between two datasets 
    SSR = 0 # sum of squares regression
    SST = 0 # sum of squares total
    mean = np.mean(data)
    for i in range(len(data)):

        SSR = SSR + (data[i] - fittedCurve[i])**2
        SST = SST + (data[i] - mean)**2

    R_squared = 1 - SSR/SST # definition of r squared
    return R_squared

# Cubic Distribution (from Ben)
def cubicDistro(x,mu_0,k,L):
    # This function is very sensitive to inputs and scipy will no be able to optimise it without good starting parameters
    # Start curve fitting with the following parameters:
    # mu_0 = 200000
    # k = 1
    # L = -0.1
    return mu_0 * (-2 * np.abs(x - 2*k*L)**3 + np.abs(x + (2-2*k)*L)**3 + np.abs(x + (-2-2*k)*L)**3 - 2*np.abs(x+2*L)**3 + 4 * np.abs(x)**3 - 2 * np.abs(x-2*L)**3 + np.abs(x + (2 + 2*k)*L)**3 + np.abs(x + (-2+2*k)*L)**3 - 2 * np.abs(x + 2*k*L)**3)

# gauss distribution
def gaussDistro(x,amplitude,mean,sigma):
    return amplitude * np.exp(-(x - mean)**2/(2*sigma**2))

def quadratic(x,a,b,c):
    return a * x**2 + b * x + c

def gauss(xcor, vals="default"):
    # get parameters

    params = loadFunctionParameters(vals,defaultValue="I:/My Drive/my-emacs/University of Bath/Year 4/ME40321/Code/simulation/gauss_parameters.npy")

    amplitude = params[0]
    mean = params[1]
    sigma = params[2]

    return gaussDistro(xcor,amplitude,mean,sigma)

def sigFig(x, sig, returnType="string"):
    # Round to sig number of significant figures
    # if a string output is desired, 0's will be added to the end to show the sig figs
    output = round(x, sig-int(np.floor(np.log10(abs(x))))-1)
    
    if returnType == "string":
        stringOutput = str(output)
        while len(stringOutput.replace(".","")) < sig:
            stringOutput = stringOutput + "0"
        return stringOutput
    elif returnType == "float":
        return output
    else:
        raise Exception("Incorrect returnType mode '" + str(returnType) +"'")

def invGauss(ycor,vals="default",):
    # Inverse of the gauss function

    # vals is either a sttring ("default") or a list of parameters for the gaussian function
    
    # get parameters    
    params = loadFunctionParameters(vals,defaultValue="I:/My Drive/my-emacs/University of Bath/Year 4/ME40321/Code/simulation/gauss_parameters.npy")
    
    amplitude = params[0]
    zeta = params[1]
    sigma = params[2]
    

    underRoot = 2*sigma**2*(np.log(amplitude) - np.log(ycor))
    if underRoot < 0:
        #print("Warning, approximating -ve square root as 0")
        underRoot = 0
        
    x1 = zeta - np.sqrt(underRoot)
    x2 = zeta + np.sqrt(underRoot)
    return [x1,x2]

def loadFunctionParameters(inputValue,defaultValue=""):
    if defaultValue == "":
        raise Exception("Default value not defined. Default value must be defined")
    
    if type(inputValue) == str:
        if inputValue == "default":
            params = np.load(defaultValue)
        else:
            filepath = "I:/My Drive/my-emacs/University of Bath/Year 4/ME40321/Code/simulation/" + str(inputValue) + ".npy"
            if os.path.exists(filepath) == False:
                raise Exception("Cannot load gauss paramters. File: " + str(filepath) + " doees not exist")
            params = np.load(filepath)
            
    else:
        params = inputValue
    return params
    
"""
# not needed anymore
def inter_element_distance_estimate(sensor_1, sensor_2,distanceVector, timeVector, datumIndex, correlationFunction,vals="default"):
    
    # CorrelationFunction is the function useed to generate the points entered into the distance estimation function
    # Vals is the gaussian approximation values to use to estimate distance
    
    inter_correlation = inter_element_correlation(sensor_1, sensor_2,distanceVector,datumIndex,correlationFunction)
    
    distancetonextsensor = sensor_2.separation - sensor_1.separation
    [estTime, estDist] = sensor_1.estimate_displacement(inter_correlation, timeVector,vals=vals)
    estDist = estDist + distancetonextsensor    
    return [estTime, estDist]
"""